"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const Types = require("../types");
const utils_1 = require("../../utils");
const render_1 = require("../render");
const constants_1 = require("../../constants");
const flatten = require("lodash.flatten");
class Transformer {
    constructor() {
        this.declarations = [];
        this.statements = [];
    }
    endPointToDeclaration(ep) {
        const nameByUrl = utils_1.getValidVarName(ep.url + '_' + ep.method).replace(/^_+/, '');
        ep.name = ep.name || nameByUrl;
        const name = ep.name;
        function toReference(type, name) {
            if (!type) {
                return new Types.TypeReferenceType(name, new Types.Any());
            }
            if (Types.isTypeReference(type)) {
                return type;
            }
            return new Types.TypeReferenceType(name, type);
        }
        const pathParams = toReference(ep.urlParams, name + '_parameter_path');
        const bodyParams = toReference(ep.bodyParams, name + '_parameter_body');
        const queryParams = toReference(ep.queryParams, name + '_parameter_query');
        const headerParams = toReference(ep.headerParams, name + '_parameter_header');
        this.declarations.push(...flatten(flatten([pathParams, bodyParams, queryParams, headerParams]).map(p => p.getDeclaration())));
        const parameters = [
            { key: 'path', type: pathParams },
            { key: 'query', type: queryParams },
            { key: 'headers', type: headerParams },
            { key: 'body', type: bodyParams },
        ];
        /** Create function below */
        const url = ts.createVariableStatement([constants_1.Export], [ts.createVariableDeclaration(name + '_url', undefined, ts.createLiteral(ep.url))]);
        const method = ts.createVariableStatement([constants_1.Export], [ts.createVariableDeclaration(name + '_method', undefined, ts.createLiteral(ep.method))]);
        this.statements.push(url, method);
        const id = (x) => ts.createIdentifier(x);
        const notEmitParameters = parameters.filter(x => x.type.isFalsy()).map(x => x.key);
        // * Not in use
        function depercationWarning() {
            if (ep.JSDoc && ep.JSDoc.depercated) {
                return ts.createStatement(ts.createCall(id('console.warn'), undefined, [ts.createLiteral(`${name} is depercated`)]));
            }
        }
        const FunctionBody = ts.createBlock([
            ts.createReturn(
            // _.request(...)
            ts.createCall(id('_.request'), undefined, [
                // name_url, name_method, { query, body, path, headers, bodyType}
                id(name + '_url'),
                id(name + '_method'),
                ts.createObjectLiteral([
                    ts.createPropertyAssignment('query', id('query')),
                    ts.createPropertyAssignment('body', id('body')),
                    ts.createPropertyAssignment('path', id('path')),
                    ts.createPropertyAssignment('headers', id('headers')),
                    ts.createPropertyAssignment('bodyType', ts.createLiteral(ep.bodyParamsType)),
                ].filter(x => -1 === notEmitParameters.indexOf(x.name.escapedText.toString()))),
            ])),
        ], true);
        function createResponse(statusCode, ref, header) {
            const TypeArguments = [statusCode, ref, header];
            if (header.kind === ts.SyntaxKind.AnyKeyword) {
                TypeArguments.pop();
                if (ref.kind === ts.SyntaxKind.AnyKeyword) {
                    TypeArguments.pop();
                }
            }
            return ts.createTypeReferenceNode('_Response', TypeArguments);
        }
        const returnTypesUnion = (result => {
            if (!ep.response) {
                return constants_1.AnyType;
            }
            const refs = ep.response
                .map(({ status: code, returnType: type, header: header }) => {
                if (type.isFalsy())
                    return null;
                if (!Types.isLiteralType(type)) {
                    const ref = new Types.TypeReferenceType(name + '_response_' + code, type);
                    this.declarations.push(...type.getDeclaration());
                    // type = ref
                }
                let headerType = new Types.Any();
                if (header) {
                    if (!Types.isTypeReference(header)) {
                        headerType = new Types.TypeReferenceType(name + '_response_header_' + code, header);
                    }
                    else {
                        headerType = header;
                    }
                    this.declarations.push(...headerType.getDeclaration());
                }
                return { code, type: type.toTypescript(), header: headerType.toTypescript() };
            })
                .filter((x) => x);
            if (refs.length === 0) {
                return createResponse(constants_1.AnyType, constants_1.AnyType, constants_1.AnyType);
            }
            if (refs.length === 1) {
                const r = refs[0];
                return createResponse(new Types.Literal(r.code, true).toTypescript(), r.type, r.header);
            }
            return ts.createUnionTypeNode(refs.map(x => createResponse(new Types.Literal(x.code, true).toTypescript(), x.type, x.header)));
        })(ep.response);
        this.declarations.push(utils_1.GenerateAsyncFunction(name, FunctionBody, parameters
            .filter(x => !x.type.isFalsy())
            .map(x => ts.createParameter(undefined, undefined, undefined, x.key, undefined, x.type.toTypescript())), returnTypesUnion, utils_1.createJSDoc(ep.JSDoc)));
        this.removeDuplicateDeclarations();
    }
    removeDuplicateDeclarations() {
        const names = [];
        this.declarations = this.declarations.filter(dec => {
            const name = ts.getNameOfDeclaration(dec).escapedText.toString();
            if (names.indexOf(name) !== -1)
                return false;
            names.push(name);
            return true;
        });
    }
}
function RestClientGenerator(server, template, config) {
    const render = new render_1.Render({ declarationOnly: config.declarationOnly });
    const transformer = new Transformer();
    server.endpoints.map(x => transformer.endPointToDeclaration(x));
    // Read data
    const statements = transformer.statements.map(render_1.Render.nodeToString).join('\n');
    const declarations = transformer.declarations.map(render_1.Render.nodeToString).join('\n');
    const result = template + '\n' + statements + '\n' + declarations;
    const file = render.createSourceFile('code.ts', result);
    file.saveSync();
    const code = render.schema2tsEmit('code.ts');
    return render_1.Render.injectTemplateVariables(config.leadingComments + '\n' + code);
}
exports.RestClientGenerator = RestClientGenerator;
//# sourceMappingURL=rest-client.js.map