import * as ts from 'typescript';
export declare enum LiteralType {
    boolean = 1,
    number = 2,
    string = 3,
}
export declare enum FalsyType {
    null = -50,
    undefined = -49,
}
export declare enum TypescriptFalsyType {
    any = -100,
    void = -99,
}
export declare enum ComplexType {
    array = 200,
    object = 201,
}
export declare enum ConstructedType {
    or = 300,
    and = 301,
    tuple = 302,
}
export declare enum TypescriptType {
    enum = 400,
    TypeReference = 401,
}
export declare function isLiteralType(x: Type): x is Literal;
export declare function isObjectType(x: Type): x is ObjectOf;
export declare function isArrayType(x: Type): x is ArrayOf;
export declare function isVoidType(x: Type): x is Void;
export declare function isTypeReference(x: Type): x is TypeReferenceType;
export declare abstract class Type {
    abstract toTypescript(): ts.TypeNode;
    /** Cut type information to be human-friendly, but maybe less precise */
    reduce(keepPrecise?: boolean): Type;
    /** To make this type referenceable, what declaration need to generate */
    getDeclaration(): ts.Declaration[];
    type: LiteralType | FalsyType | ComplexType | ConstructedType | TypescriptFalsyType | TypescriptType;
    isFalsy(this: Type): boolean;
    protected JSDoc: string | undefined;
    protected JSDocType: ts.SyntaxKind.SingleLineCommentTrivia | ts.SyntaxKind.MultiLineCommentTrivia;
    addJSDoc(jsdoc: string | string[], type?: ts.SyntaxKind.SingleLineCommentTrivia | ts.SyntaxKind.MultiLineCommentTrivia): void;
    JSDocWrapper<T extends ts.Node>(node: T, newLine?: boolean): T;
}
export declare class Literal extends Type {
    value: boolean | number | string | undefined | null;
    /** Does this type have a literal value? */ literal: boolean;
    constructor(value: boolean | number | string | undefined | null, 
        /** Does this type have a literal value? */ literal: boolean);
    toTypescript(): ts.KeywordTypeNode | ts.LiteralTypeNode;
    reduce(keepPrecise?: boolean): Type;
}
export declare class Any extends Type {
    type: TypescriptFalsyType;
    toTypescript(): ts.KeywordTypeNode;
}
export declare class Void extends Type {
    type: TypescriptFalsyType;
    toTypescript(): ts.KeywordTypeNode;
}
export declare class ArrayOf<T extends Type = Type> extends Type {
    of: T;
    type: ComplexType;
    constructor(of: T);
    getDeclaration(): ts.Declaration[];
    toTypescript(): ts.ArrayTypeNode;
    reduce(): ArrayOf<Type>;
}
export interface ObjectOfWhat {
    key: string;
    optional?: boolean;
    value: Type;
    jsdoc?: string;
    readonly?: boolean;
    defaultValue?: any;
}
export declare class ObjectOf extends Type {
    of: ObjectOfWhat[];
    type: ComplexType;
    constructor(of: ObjectOfWhat[]);
    toTypescript(): ts.TypeLiteralNode;
    reduce(): ObjectOf;
    getDeclaration(): ts.Declaration[];
}
export declare class EnumOf extends Type {
    name: string;
    of: {
        name: string;
        value?: string | number | undefined;
    }[];
    type: TypescriptType;
    constructor(name: string, of?: {
        name: string;
        value?: string | number | undefined;
    }[]);
    getDeclaration(): ts.Declaration[];
    toTypescript(): ts.TypeReferenceNode;
}
export declare class TypeReferenceType extends Type {
    ref: string;
    of: Type;
    type: TypescriptType;
    constructor(ref: string, of: Type);
    toTypescript(): ts.KeywordTypeNode | ts.TypeReferenceNode | ts.ArrayTypeNode | ts.LiteralTypeNode;
    getDeclaration(): ts.Declaration[];
}
export declare class Or extends Type {
    of: Type[];
    type: ConstructedType;
    constructor(of: Type[]);
    toTypescript(): ts.UnionTypeNode;
    reduce(): Type;
    getDeclaration(): ts.Declaration[];
}
export declare class And extends Type {
    of: Type[];
    type: ConstructedType;
    constructor(of: Type[]);
    toTypescript(): ts.IntersectionTypeNode;
    reduce(): Type;
    getDeclaration(): ts.Declaration[];
}
export declare class TupleOf extends Type {
    of: Type[];
    type: ConstructedType;
    constructor(of: Type[]);
    toTypescript(): ts.TupleTypeNode;
    reduce(): TupleOf;
    getDeclaration(): ts.Declaration[];
}
/** Geneate a shape of the object
 * if guessMode is on, we will generate a more grace but less correctness type for it
 */
export declare function shape(any: any, guessMode?: boolean): Type;
