"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
//#region Types & fns
var LiteralType;
(function (LiteralType) {
    LiteralType[LiteralType["boolean"] = 1] = "boolean";
    LiteralType[LiteralType["number"] = 2] = "number";
    LiteralType[LiteralType["string"] = 3] = "string";
})(LiteralType = exports.LiteralType || (exports.LiteralType = {}));
var FalsyType;
(function (FalsyType) {
    FalsyType[FalsyType["null"] = -50] = "null";
    FalsyType[FalsyType["undefined"] = -49] = "undefined";
})(FalsyType = exports.FalsyType || (exports.FalsyType = {}));
var TypescriptFalsyType;
(function (TypescriptFalsyType) {
    TypescriptFalsyType[TypescriptFalsyType["any"] = -100] = "any";
    TypescriptFalsyType[TypescriptFalsyType["void"] = -99] = "void";
})(TypescriptFalsyType = exports.TypescriptFalsyType || (exports.TypescriptFalsyType = {}));
var ComplexType;
(function (ComplexType) {
    ComplexType[ComplexType["array"] = 200] = "array";
    ComplexType[ComplexType["object"] = 201] = "object";
})(ComplexType = exports.ComplexType || (exports.ComplexType = {}));
var ConstructedType;
(function (ConstructedType) {
    ConstructedType[ConstructedType["or"] = 300] = "or";
    ConstructedType[ConstructedType["and"] = 301] = "and";
    ConstructedType[ConstructedType["tuple"] = 302] = "tuple";
})(ConstructedType = exports.ConstructedType || (exports.ConstructedType = {}));
var TypescriptType;
(function (TypescriptType) {
    TypescriptType[TypescriptType["enum"] = 400] = "enum";
    TypescriptType[TypescriptType["TypeReference"] = 401] = "TypeReference";
})(TypescriptType = exports.TypescriptType || (exports.TypescriptType = {}));
function isLiteralType(x) {
    return [LiteralType.boolean, LiteralType.number, LiteralType.string].some(y => x.type === y);
}
exports.isLiteralType = isLiteralType;
function isObjectType(x) {
    return x.type === ComplexType.object;
}
exports.isObjectType = isObjectType;
function isArrayType(x) {
    return x.type === ComplexType.array;
}
exports.isArrayType = isArrayType;
function isVoidType(x) {
    return x.type === TypescriptFalsyType.void;
}
exports.isVoidType = isVoidType;
function isTypeReference(x) {
    return x.type === TypescriptType.TypeReference;
}
exports.isTypeReference = isTypeReference;
//#endregion
class Type {
    constructor() {
        this.JSDoc = undefined;
    }
    /** Cut type information to be human-friendly, but maybe less precise */
    reduce(keepPrecise) {
        return this;
    }
    /** To make this type referenceable, what declaration need to generate */
    getDeclaration() {
        return [];
    }
    isFalsy() {
        switch (this.type) {
            case TypescriptType.TypeReference:
                return this.of.isFalsy();
            case ComplexType.object:
                return this.of.length === 0;
            default:
                const falsy = {
                    [FalsyType.null]: 1,
                    [FalsyType.undefined]: 1,
                    [TypescriptFalsyType.any]: 1,
                    [TypescriptFalsyType.void]: 1,
                };
                return !!falsy[this.type];
        }
    }
    addJSDoc(jsdoc, type = ts.SyntaxKind
            .MultiLineCommentTrivia) {
        if (Array.isArray(jsdoc)) {
            jsdoc = jsdoc.join('\n* ');
        }
        this.JSDoc = jsdoc;
        this.JSDocType = type;
    }
    JSDocWrapper(node, newLine = false) {
        if (this.JSDoc)
            ts.addSyntheticLeadingComment(node, this.JSDocType, `* ${this.JSDoc} `, newLine);
        return node;
    }
}
exports.Type = Type;
//#region Literal, Any, Void
class Literal extends Type {
    constructor(value, 
        /** Does this type have a literal value? */ literal) {
        super();
        this.value = value;
        this.literal = literal;
        this.type = (LiteralType[typeof value] || FalsyType[typeof value] || FalsyType.null);
    }
    toTypescript() {
        if (this.value === undefined)
            return ts.createKeywordTypeNode(ts.SyntaxKind.UndefinedKeyword);
        if (this.value === null)
            return ts.createKeywordTypeNode(ts.SyntaxKind.NullKeyword);
        if (this.literal) {
            return ts.createLiteralTypeNode(ts.createLiteral(this.value));
        }
        const keywords = {
            boolean: ts.SyntaxKind.BooleanKeyword,
            string: ts.SyntaxKind.StringKeyword,
            number: ts.SyntaxKind.NumberKeyword,
        };
        return ts.createKeywordTypeNode(keywords[typeof this.value]);
    }
    reduce(keepPrecise) {
        /** We will drop all literals here */
        if (this.literal && !keepPrecise) {
            return shape(this.value);
        }
        else
            return this;
    }
}
exports.Literal = Literal;
class Any extends Type {
    constructor() {
        super(...arguments);
        this.type = TypescriptFalsyType.any;
    }
    toTypescript() {
        return ts.createKeywordTypeNode(ts.SyntaxKind.AnyKeyword);
    }
}
exports.Any = Any;
class Void extends Type {
    constructor() {
        super(...arguments);
        this.type = TypescriptFalsyType.void;
    }
    toTypescript() {
        return ts.createKeywordTypeNode(ts.SyntaxKind.VoidKeyword);
    }
}
exports.Void = Void;
//#endregion
//#region ComplexType
class ArrayOf extends Type {
    constructor(of) {
        super();
        this.of = of;
        this.type = ComplexType.array;
    }
    getDeclaration() {
        return this.of.getDeclaration();
    }
    toTypescript() {
        return ts.createArrayTypeNode(this.of.toTypescript());
    }
    reduce() {
        return new ArrayOf(this.of.reduce());
    }
}
exports.ArrayOf = ArrayOf;
class ObjectOf extends Type {
    constructor(of) {
        super();
        this.of = of;
        this.type = ComplexType.object;
    }
    toTypescript() {
        return ts.createTypeLiteralNode(this.of.map(x => {
            if (!x.key) {
                throw new TypeError('An unknown property name.');
            }
            const node = ts.createPropertySignature(undefined, 
            /** name */ ts.createLiteral(x.key), 
            /** question mark */ x.optional ? ts.createToken(ts.SyntaxKind.QuestionToken) : undefined, 
            /** subtype */ x.value.toTypescript(), undefined);
            x.value.JSDocWrapper(node, true);
            return node;
        }));
    }
    reduce() {
        return new ObjectOf(this.of.map(x => {
            return Object.assign({}, x, { value: x.value.reduce() });
        }));
    }
    getDeclaration() {
        const x = [];
        this.of.forEach(_ => x.push(..._.value.getDeclaration()));
        return x;
    }
}
exports.ObjectOf = ObjectOf;
//#endregion
//#region Typescript Type
class EnumOf extends Type {
    constructor(name, of = []) {
        super();
        this.name = name;
        this.of = of;
        this.type = TypescriptType.enum;
    }
    getDeclaration() {
        const members = [];
        for (const item of this.of) {
            const member = ts.createEnumMember(item.name, item.value !== undefined ? ts.createLiteral(item.value) : undefined);
            members.push(member);
        }
        const self = ts.createEnumDeclaration(undefined, [ts.createToken(ts.SyntaxKind.ExportKeyword)], this.name, members);
        return [this.JSDocWrapper(self)];
    }
    toTypescript() {
        return ts.createTypeReferenceNode(this.name, []);
    }
}
exports.EnumOf = EnumOf;
class TypeReferenceType extends Type {
    constructor(ref, of) {
        super();
        this.ref = ref;
        this.of = of;
        this.type = TypescriptType.TypeReference;
    }
    toTypescript() {
        if (this.isFalsy()) {
            if (isObjectType(this.of))
                return new Void().toTypescript();
            if (isArrayType(this.of))
                return ts.createArrayTypeNode(new Void().toTypescript());
            if (isVoidType(this.of))
                return new Void().toTypescript();
            if (isLiteralType(this.of))
                return this.of.toTypescript();
            return new Any().toTypescript();
        }
        return this.JSDocWrapper(ts.createTypeReferenceNode(this.ref, []));
    }
    getDeclaration() {
        if (this.isFalsy()) {
            return [];
        }
        /** If this.of is eliminated, do not generate type reference
         *  if this.of is a Object, generate an interface
         *  otherwise, generate a type alias (type X = Type)
         */
        let d = null;
        if (isTypeReference(this.of)) {
            return this.of.getDeclaration();
        }
        if (isObjectType(this.of)) {
            const getName = (x) => {
                if (!x.name)
                    return undefined;
                if (ts.isIdentifier(x.name) || ts.isStringLiteral(x.name)) {
                    return x.name.text;
                }
                return undefined;
            };
            const getComment = (y) => {
                const name = getName(y);
                if (!name)
                    return;
                this.of;
            };
            const nodes = this.of.toTypescript().members;
            nodes
                .map(node => ({ name: getName(node), node }))
                .filter(x => x.name)
                .map(x => {
                return Object.assign({ comment: this.of.of.filter(key => key.key === x.name)[0] }, x);
            })
                .filter(x => x.comment && x.comment.jsdoc)
                .forEach(x => {
                ts.addSyntheticLeadingComment(x.node, ts.SyntaxKind.MultiLineCommentTrivia, '* ' + x.comment.jsdoc + ' ');
            });
            d = ts.createInterfaceDeclaration(void 0, [ts.createToken(ts.SyntaxKind.ExportKeyword)], this.ref, void 0, void 0, nodes);
        }
        else {
            d = ts.createTypeAliasDeclaration(undefined, [ts.createToken(ts.SyntaxKind.ExportKeyword)], this.ref, [], this.of.toTypescript());
        }
        return [...this.of.getDeclaration(), d];
    }
}
exports.TypeReferenceType = TypeReferenceType;
//#endregion
//#region ConstructedType Helper
function RemoveDuplicateBaseType(x) {
    const newTypes = [];
    x.forEach(o => {
        if (FalsyType[o.type] ||
            (LiteralType[o.type] && o.literal === false) ||
            TypescriptFalsyType[o.type]) {
            if (newTypes.some(x => x.type === o.type))
                return;
        }
        newTypes.push(o);
    });
    return newTypes.filter(x => !FalsyType[x.type] && !TypescriptFalsyType[x.type]);
}
function CombineObjectType(x) {
    const otherTypes = x.filter(y => y.type !== ComplexType.object);
    const allObjects = x.filter(y => y.type === ComplexType.object);
    const reducedKeys = [];
    const keys = [];
    allObjects.map(x => x.of).forEach(x => keys.push(...x));
    /** We guess it is a dictionary if it has more than 20 keys
     *  so we set their keys to [key: string]
     *  this will generate like this: { [key: string]: type }
     */
    if (keys.length > 20) {
        keys.forEach(x => (x.key = '[key: string]'));
    }
    for (const key of keys) {
        const loadedKey = reducedKeys.filter(x => x.key === key.key);
        if (loadedKey.length) {
            const it = loadedKey[0];
            it.value = new Or([it.value, key.value]).reduce();
            it.jsdoc += '\n' + key.jsdoc;
            it.optional = it.optional || key.optional;
        }
        else {
            reducedKeys.push(key);
        }
    }
    const result = MakeFalsyTypeToAny(reducedKeys);
    if (result.length)
        return [...otherTypes, new ObjectOf(result)];
    return otherTypes;
    function MakeFalsyTypeToAny(o) {
        return o.map(x => {
            const more = { optional: true, value: new Any() };
            if (x.value.type === FalsyType.null || x.value.type === FalsyType.undefined)
                return Object.assign({}, x, more);
            return x;
        });
    }
}
const ReduceForConstructedType = (x) => CombineObjectType(RemoveDuplicateBaseType(
// after this reduce, something like 1 | 3 will become number | number
// so we need to RemoveDuplicateBaseType again
x.map(x => x.reduce())));
function GetFlattedAndOr(self, x) {
    const result = x.filter(y => y.type !== self.type);
    const others = x.filter(y => y.type === self.type);
    if (others.length)
        return result.concat(...others.map(x => GetFlattedAndOr(self, x.of)));
    return result;
}
//#endregion
//#region ConstructedType
class Or extends Type {
    constructor(of) {
        super();
        this.of = of;
        this.type = ConstructedType.or;
        this.of = RemoveDuplicateBaseType(this.of);
    }
    toTypescript() {
        return ts.createUnionTypeNode(this.of.map(x => x.toTypescript()));
    }
    reduce() {
        const r = GetFlattedAndOr(this, ReduceForConstructedType(this.of));
        if (r.length === 0)
            return new Any();
        if (r.length === 1)
            return r[0];
        return new Or(r);
    }
    getDeclaration() {
        const x = [];
        this.of.forEach(_ => x.push(..._.getDeclaration()));
        return x;
    }
}
exports.Or = Or;
class And extends Type {
    constructor(of) {
        super();
        this.of = of;
        this.type = ConstructedType.and;
        this.of = RemoveDuplicateBaseType(of);
    }
    toTypescript() {
        return ts.createIntersectionTypeNode(this.of.map(x => x.toTypescript()));
    }
    reduce() {
        const r = GetFlattedAndOr(this, ReduceForConstructedType(this.of));
        if (r.length === 0)
            return new Any();
        if (r.length === 1)
            return r[0];
        return new And(r);
    }
    getDeclaration() {
        const x = [];
        this.of.forEach(_ => x.push(..._.getDeclaration()));
        return x;
    }
}
exports.And = And;
class TupleOf extends Type {
    constructor(of) {
        super();
        this.of = of;
        this.type = ConstructedType.tuple;
    }
    toTypescript() {
        return ts.createTupleTypeNode(this.of.map(x => x.toTypescript()));
    }
    reduce() {
        return new TupleOf(this.of.map(x => x.reduce()));
    }
    getDeclaration() {
        const x = [];
        this.of.forEach(_ => x.push(..._.getDeclaration()));
        return x;
    }
}
exports.TupleOf = TupleOf;
//#endregion
/** Geneate a shape of the object
 * if guessMode is on, we will generate a more grace but less correctness type for it
 */
function shape(any, guessMode) {
    let returnValue;
    if (any instanceof Type) {
        returnValue = any;
    }
    else if (['boolean', 'number', 'string', 'undefined'].indexOf(typeof any) + 1 || any === null) {
        returnValue = new Literal(any, false);
    }
    else if (Array.isArray(any)) {
        returnValue = new ArrayOf(new Or(any.map(x => shape(x))));
    }
    else
        returnValue = new ObjectOf(Object.keys(any).map(key => {
            return {
                key: key,
                value: shape(any[key]),
                optional: any[key] === undefined || any[key] === null,
            };
        }));
    if (guessMode === true)
        return returnValue.reduce();
    return returnValue;
}
exports.shape = shape;
//# sourceMappingURL=types.js.map