"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Swagger2 = require("swagger-parser");
const Types = require("../../../code/types");
const utils_1 = require("../../../utils");
const cloneDeep = require("lodash.clonedeep");
function is(object) {
    if (object.swagger != '2.0') {
        return false;
    }
    // validate isn't a sync function, so we let it go through, if validate failed, throw then
    // !?!?! validate dereferenced JSON $ref?????
    Swagger2.validate(cloneDeep(object)).catch(reason => {
        throw reason;
    });
    return true;
}
exports.is = is;
const baseTypeMap = {
    string: '',
    boolean: true,
    integer: 1,
    number: 1,
};
async function transformer(doc) {
    const JSONSchemaToTypes = await utils_1.createJSONSchemaToTypes(doc, true);
    const baseUrl = doc.basePath;
    const endpoints = [];
    const enums = [];
    const interfaces = [];
    for (const pathKey in doc.paths) {
        const path = doc.paths[pathKey];
        generateEndpoint(pathKey, 'get', path.get);
        generateEndpoint(pathKey, 'post', path.post);
        generateEndpoint(pathKey, 'put', path.put);
        generateEndpoint(pathKey, 'delete', path.delete);
        generateEndpoint(pathKey, 'head', path.head);
        generateEndpoint(pathKey, 'options', path.options);
        generateEndpoint(pathKey, 'patch', path.patch);
    }
    return {
        baseUrl: baseUrl || '/',
        endpoints,
        enums,
        interfaces,
    };
    function generateEndpoint(path, method, op) {
        if (!op)
            return;
        const JSDoc = {
            depercated: op.deprecated,
            description: op.description || (op.externalDocs && op.externalDocs.description),
            summary: op.summary,
            see: op.externalDocs ? op.externalDocs.url : undefined,
            example: undefined,
        };
        let headerParams, queryParams, pathParams, formParams, bodyParams;
        const result = [];
        if (op.parameters) {
            function fromParameter(of) {
                const n = op.parameters.filter(x => x.in === of);
                if (n.length === 0) {
                    return undefined;
                }
                return new Types.ObjectOf(n.map(_ => {
                    // Also works with QueryParameter, HeaderParameter, PathParameter, FormDataParameter
                    const query = _;
                    const schema = Object.assign({}, query, { required: undefined });
                    return {
                        key: query.name,
                        value: JSONSchemaToTypes(schema),
                        defaultValue: query.default,
                        jsdoc: query.description,
                    };
                }));
            }
            queryParams = fromParameter('query');
            headerParams = fromParameter('header');
            pathParams = fromParameter('path');
            formParams = fromParameter('formData');
            op.parameters.filter(x => x.in === 'body').forEach((body) => {
                const t = JSONSchemaToTypes(body.schema);
                t.optional = !body.required;
                bodyParams = t;
            });
        }
        let bodyParamsType = 'json';
        if (op.parameters) {
            bodyParamsType = op.parameters.some(p => p.in === 'formData') ? 'form' : 'json';
        }
        if (op.responses) {
            for (const status in op.responses) {
                const num = parseInt(status, 10);
                if (status === 'default' || !isNaN(num)) {
                    const r = op.responses[status];
                    result.push({
                        status: !isNaN(num) ? num : status,
                        returnType: r.schema ? JSONSchemaToTypes(r.schema) : new Types.Any(),
                        header: r.headers &&
                            JSONSchemaToTypes({ type: 'object', properties: r.headers }),
                    });
                }
            }
        }
        endpoints.push({
            JSDoc: JSDoc,
            url: path,
            urlParams: pathParams,
            headerParams,
            queryParams,
            response: result,
            name: op.operationId ? utils_1.getValidVarName(op.operationId) : undefined,
            bodyParamsType,
            bodyParams: bodyParamsType === 'json' ? bodyParams : formParams,
            method: method,
            modifier: {},
        });
    }
}
exports.transformer = transformer;
//# sourceMappingURL=index.js.map