"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const yaml = require("yamljs");
const ts = require("typescript");
const url_1 = require("url");
const request = require("request");
const constants_1 = require("./constants");
const fs_1 = require("fs");
const path_1 = require("path");
const fileSystem_1 = require("ts-simple-ast/dist/fileSystem");
const Types = require("./code/types");
const $RefParser = require("json-schema-ref-parser");
// tslint:disable-next-line:quotemark (conflict with prettier)
const ValidCharsInURLSpecButNOTInVarName = "-._~:/?#[]@!&'()*+,;= ".split('');
/**
 * A FileSystemHost provided to typescript compiler
 */
class ReadonlyFileSystemHost extends fileSystem_1.DefaultFileSystemHost {
    constructor() {
        super(...arguments);
        this.changes = new Map();
    }
    getPath(filePath) {
        return path_1.normalize(path_1.join(this.getCurrentDirectory(), filePath));
    }
    async writeFile(filePath, fileText) {
        this.changes.set(this.getPath(filePath), fileText);
    }
    writeFileSync(filePath, fileText) {
        this.changes.set(this.getPath(filePath), fileText);
    }
    readFileSync(filePath, encoding) {
        return this.changes.get(this.getPath(filePath)) || super.readFileSync(filePath, encoding);
    }
    async readFile(filePath, encoding) {
        const f = this.changes.get(this.getPath(filePath));
        return f ? Promise.resolve(f) : super.readFile(filePath, encoding);
    }
}
exports.ReadonlyFileSystemHost = ReadonlyFileSystemHost;
/** Get a valid variable name from url string */
function getValidVarName(name) {
    return name
        .split('')
        .map(char => (ValidCharsInURLSpecButNOTInVarName.indexOf(char) !== -1 ? '_' : char))
        .join('')
        .replace('{', '$')
        .replace('}', '');
}
exports.getValidVarName = getValidVarName;
/** Parse JSON or Yaml */
function parseJSONorYAML(text) {
    try {
        return JSON.parse(text);
    }
    catch (jsonerr) {
        try {
            return yaml.parse(text);
        }
        catch (yamlerr) {
            const jsonmsg = jsonerr.message;
            const yamlmsg = yamlerr.message;
            throw new SyntaxError(`JSON and Yaml parser failed to parse the input file
  JSON: ${jsonmsg}
  Yaml: ${yamlmsg} at line ${yamlerr.parsedLine || 'unknown'}`);
        }
    }
}
exports.parseJSONorYAML = parseJSONorYAML;
/** Request a file */
function requestFile(url) {
    try {
        const u = new url_1.URL(url);
        // This also includes https
        if (!u.protocol.includes('http')) {
            throw new Error('This is a file');
        }
        return new Promise((resolve, reject) => {
            request(url, (error, response, body) => {
                if (error)
                    return reject(error);
                try {
                    return resolve(body);
                }
                catch (e) {
                    return reject(e);
                }
            });
        });
    }
    catch (_a) {
        return Promise.resolve(fs_1.readFileSync(url, 'utf-8'));
    }
}
exports.requestFile = requestFile;
/** Create a Typescript Async function Declaration */
function GenerateAsyncFunction(name, body, parameters = [], returnType = constants_1.AnyType, JSDocCommet, modifiers = []) {
    const returnTypePromise = ts.createTypeReferenceNode('Promise', [returnType]);
    const node = ts.createFunctionDeclaration(undefined, [constants_1.Export, ...modifiers], undefined, name, undefined, parameters, returnTypePromise, body);
    JSDocCommet && ts.addSyntheticLeadingComment(node, ts.SyntaxKind.MultiLineCommentTrivia, `* ${JSDocCommet} `, true);
    return node;
}
exports.GenerateAsyncFunction = GenerateAsyncFunction;
/** JSON Schema -> Schema2ts inner express */
async function createJSONSchemaToTypes(document, nameOptions) {
    const doc = await $RefParser.resolve(document);
    return function JSONSchemaToTypes(from) {
        //#region Speical properties
        if (from.$ref) {
            const result = doc.get(from.$ref);
            const name = exports.getJSONRefName(from, nameOptions);
            try {
                return new Types.TypeReferenceType(name, JSONSchemaToTypes(result));
            }
            catch (_a) {
                return new Types.TypeReferenceType(name, Types.shape(result));
            }
        }
        else if (from.allOf) {
            /**
             * allOf, anyOf, oneOf, See this
             * https://tools.ietf.org/html/draft-fge-json-schema-validation-00#section-5.5.3 */
            return new Types.And(from.allOf.map(x => JSONSchemaToTypes(x)));
        }
        else if (from.anyOf) {
            console.warn(`Typescript can not express JSON 'anyOf' easily, treat as 'oneOf'`);
            const newer = Object.assign({}, from, { oneOf: from.anyOf, anyOf: undefined });
            return JSONSchemaToTypes(newer);
        }
        else if (from.oneOf) {
            return new Types.Or(from.oneOf.map(x => JSONSchemaToTypes(x)));
        }
        else if (from.additionalProperties) {
            const newer = Object.assign({}, from, { additionalProperties: undefined });
            return new Types.And([JSONSchemaToTypes(newer), JSONSchemaToTypes(from.additionalProperties)]);
        }
        else if (from.enum) {
            const newer = Object.assign({}, from, { enum: undefined });
            if (from.type === 'string' || from.type === 'number') {
                const enums = from.enum.filter(x => typeof x === 'string');
                let EnumName = undefined;
                // Speical case for OpenAPI 2.0
                if (document.definitions) {
                    for (const key in document.definitions) {
                        const _def = document.definitions[key].enum;
                        // Check if enum in the definitions is the same as current enum
                        // Cause OpenAPI 2.0 enum use yaml reference, so the name is lost in the convertion
                        // We have to use this way to compare
                        if (_def &&
                            document.definitions[key].type === from.type && // string enum is not equal to number enum
                            Array.isArray(_def) &&
                            _def.every((v, i) => v === from.enum[i]) // Since the enum ref is using yaml ref, it must keep the same order
                        ) {
                            EnumName = key;
                        }
                    }
                }
                if (EnumName) {
                    // We get the name of enum
                    return new Types.EnumOf(EnumName, enums.map((x, index) => ({
                        name: x,
                        value: from.type === 'string' ? x : index,
                    })));
                }
                else {
                    // We can't figure out the name of enum, so we have to use a | b | c type
                    const or = new Types.Or(enums.map((member, index) => new Types.Literal(from.type === 'string' ? member : index, true)));
                    from.type === 'number' && or.addJSDoc(['enum:', ...enums.map((x, index) => `  ${x}: ${index}`)]);
                    return or;
                }
            }
            else {
                console.warn(`Schema2ts can not create **enum** of type **${from.type}**, use it original type instead`);
                return JSONSchemaToTypes(newer);
            }
        }
        //#endregion
        const type = from && from.type;
        if (isNumber(type))
            return new Types.Literal(0, false);
        switch (type) {
            case 'array':
                if (Array.isArray(from.items)) {
                    return new Types.ArrayOf(JSONSchemaToTypes(from.items[0]));
                }
                else if (!from.items) {
                    return new Types.Any();
                }
                else {
                    return new Types.ArrayOf(JSONSchemaToTypes(from.items));
                }
            case 'object':
                const props = from.properties;
                const ofWhat = [];
                for (const key in props) {
                    const schema = props[key];
                    ofWhat.push({
                        key: key,
                        value: JSONSchemaToTypes(schema),
                        optional: (from.required || []).every(x => x !== key),
                        jsdoc: from.description,
                        readonly: from.readOnly,
                        defaultValue: from.default,
                    });
                }
                return new Types.ObjectOf(ofWhat);
            case 'string':
                return new Types.Literal('', false);
            case 'boolean':
                return new Types.Literal(true, false);
            case 'null':
                return new Types.Any();
            default:
                try {
                    const newer = Object.assign({}, from, { type: 'object' });
                    return JSONSchemaToTypes(newer);
                }
                catch (_b) {
                    return JSONSchemaToTypes({ type: 'null' });
                }
        }
    };
}
exports.createJSONSchemaToTypes = createJSONSchemaToTypes;
function isNumber(n) {
    switch (n) {
        case 'integer':
        case 'long':
        case 'float':
        case 'double':
        case 'number':
            return true;
        default:
            return false;
    }
}
exports.getJSONRefName = (x, openapi) => {
    // For valid name, see https://tools.ietf.org/html/draft-ietf-appsawg-json-pointer-04#section-3
    let name = x.$ref;
    if (typeof name !== 'string')
        return undefined;
    if (name === '#') {
        return 'Document';
    }
    if (name.startsWith('#'))
        name = name.replace(/^#/, '');
    if (name === '/')
        return 'Document';
    if (name.startsWith('/'))
        name = name.replace(/^\//, '');
    if (openapi) {
        if (name.startsWith('definitions'))
            return getValidVarName(name).replace(/^definitions_/, '');
    }
    return getValidVarName(name);
};
function createJSDoc(jsdoc) {
    if (!jsdoc || Object.keys(jsdoc).filter(x => jsdoc[x] !== undefined).length === 0) {
        return;
    }
    let comment = '';
    if (jsdoc.depercated)
        comment += '\n@depercated';
    if (jsdoc.summary)
        comment += '\n@summary ' + jsdoc.summary;
    if (jsdoc.description)
        comment += '\n@description ' + jsdoc.description;
    if (jsdoc.see)
        comment += '\n@see ' + jsdoc.see;
    comment = comment.replace(/^\n/, '');
    if (comment.match(/\n/))
        comment += '\n';
    return comment;
}
exports.createJSDoc = createJSDoc;
//# sourceMappingURL=utils.js.map