export { RestAPI as Schema2tsServerDefinition } from './code/server'
import schema2server from './specifications'

import { readFileSync } from 'fs'
import { join } from 'path'
import { requestFile, parseJSONorYAML } from './utils'
import { RestClientGenerator } from './code/generator/rest-client'

export interface Schema2tsAPI {
    /** Custom template that used to generate code */ template?: string
    /** If this is true, template will be treated as url/file path */ isTemplateUrl?: boolean
    /** Schema that used to generate code */ schema: string | object
    /** If this is true, schema will be treated as url/file path */ isSchemaUrl?: boolean
    /** Generate only declarations */ declaration?: boolean
    /** If you only want to change comments on the top, you may need this. */ customFileComment?: string
    /** Code generator */ generator?: (server: any, template: string, config: any) => string
}

const defaultTemplate = join(__dirname, '..', 'src', 'default.template.ts')
const defaultTemplateComment = `*  This file is auto-generated on %when%. DO NOT modify this file directly!
*  Generated by:
* 	    schema2ts %version%
* 	    Typescript %typescript-version%
*
*  Schema2ts: https://github.com/Jack-Works/schema2tscode/
*  Syntax Error in this file? Report there https://github.com/Jack-Works/schema2tscode/issues`

export default async function({
    isTemplateUrl,
    template = isTemplateUrl ? defaultTemplate : readFileSync(defaultTemplate, 'utf-8'),
    schema,
    customFileComment = defaultTemplateComment,
    generator = RestClientGenerator as Schema2tsAPI['generator'],
    ...config
}: Schema2tsAPI) {
    if (isTemplateUrl) {
        template = await requestFile(template)
    }
    if (!schema) {
        throw new ReferenceError('No input file, use --schema [url/file path] to provide one')
    }
    if (config.isSchemaUrl && typeof schema === 'string') {
        schema = parseJSONorYAML(await requestFile(schema))
    }
    if (typeof schema === 'string') {
        schema = parseJSONorYAML(schema)
    }
    // TODO: Prevent security problem caused by this.
    const leadingComments = `/* tslint:disable */
/*--------------------------------------------------------------------------------------------
${customFileComment.replace('%default-template%', defaultTemplateComment)}
*--------------------------------------------------------------------------------------------*/`
    const internalExpressOfSchema = await schema2server(schema as object)
    const code = generator!(internalExpressOfSchema, template, {
        declarationOnly: config.declaration,
        leadingComments: leadingComments,
    })
    return code
}
